{ *****************************************************
              TDBFilter Component

  TDBFilter is a component to make it easy to filter
  standard tables on non-indexed fields. This gives the
  user much greater control of the data. It is also very
  useful when you are forced to index on a key field
  (such as in a master detail relationship) but you
  need to set a range on another field.

                  Paul Warren
         HomeGrown Software Development
       (c) 1997 Langley British Columbia.
                (604) 856-6523
         e-mail:  hg_soft@uniserve.com
    Home page: http://users.uniserve.com/~hg_soft
  ***************************************************** }

unit Dbfilter;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, {$IFDEF WIN32} DBCtrls, DBCommon,{$ENDIF} DBTables,
  DBiTypes, DBiProcs, DBiErrs;

type
  TOpType = (optEQ, optNE, optGT, optLT, optGE, optLE);

  TDBFilter = class(TComponent)
  private
    { Private declarations }
    FDataLink: TFieldDataLink;
    FFilter: string;
    FApply: boolean;
    FOperator: TOpType;
    Buf: array[0..100] of char;
    function GetField: TField;
    procedure SetFilter(value: string);
    procedure ApplyFilter(value: boolean);
    function GetDataSource: TDataSource;
    function GetDataField: string;
    procedure SetDataSource(Value: TDataSource);
    procedure SetDataField(const Value: string);
    procedure FilterToBuf;
  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    procedure AddFilter;
    procedure RemoveFilter;
    property Field: TField read GetField;
  published
    { Published declarations }
    property Apply: boolean read FApply write ApplyFilter default false;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DataField: string read GetDataField write SetDataField;
    property Filter: string read FFilter write SetFilter;
    property Operator: TOpType read FOperator write FOperator default optEQ;
  end;

{$IFNDEF Win32}
const
  FldTypeMap: array[TFieldType] of Byte = (
    fldUNKNOWN, fldZSTRING, fldINT16, fldINT32, fldUINT16, fldBOOL,
    fldFLOAT, fldFLOAT, fldBCD, fldDATE, fldTIME, fldTIMESTAMP, fldBYTES,
    fldVARBYTES, fldINT32, fldBLOB, fldBLOB);
{$ENDIF}

const
  OpTypeMap: array[TOpType] of CanOP = (canEQ, canNE, canGT, canLT, canGE, canLE);

procedure Register;

implementation

{$IFDEF WIN32}
  {$R DBFILTER.D32}
{$ELSE}
  {$R DBFILTER.D16}
{$ENDIF}

constructor TDBFilter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  { create the datalink }
  FDataLink := TFieldDataLink.Create;
  { zero the buffer }
  FillChar(Buf, SizeOf(Buf), #0);
  FOperator := optEQ;
  FFilter := '';
  FApply := false;
end;

destructor TDBFilter.Destroy;
begin
  { free the datalink }
  FDataLink.Free;
  { make it explicitly nil }
  FDataLink := nil;
  inherited Destroy;
end;

procedure TDBFilter.Loaded;
begin
  inherited Loaded;	{ always call the inherited Loaded first! }
  FilterToBuf;
end;

{ method to get DataSource from DataLink }
function TDBFilter.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

{ method to set DataSource in DataLink }
procedure TDBFilter.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

{ method to get DataField }
function TDBFilter.GetDataField: string;
begin
  Result := FDataLink.FieldName
end;

{ method to set DataField }
procedure TDBFilter.SetDataField(const Value: string);
const
{ Comment out next section for < Delphi 3 }
{$IFDEF WIN32}
  FldType: array[TFieldType] of string = ('ftUnknown', 'ftString', 'ftSmallint',
  'ftInteger', 'ftWord', 'ftBoolean', 'ftFloat', 'ftCurrency', 'ftBCD', 'ftDate',
  'ftTime', 'ftDateTime', 'ftBytes', 'ftVarBytes', 'ftAutoInc', 'ftBlob', 'ftMemo',
  'ftGraphic', 'ftFmtMemo', 'ftParadoxOle', 'ftDBaseOle', 'ftTypedBinary', 'ftCursor',
  'ftFixedChar', 'ftWideString', 'ftLargeint', 'ftADT', 'ftArray', 'ftReference',
  'ftDataSet');

(******* Remove these comments for < Delphi 3 {$IFDEF WIN32}
  FldType: array[TFieldType] of string = ('ftUnknown', 'ftString', 'ftSmallint',
  'ftInteger', 'ftWord', 'ftBoolean', 'ftFloat', 'ftCurrency', 'ftBCD', 'ftDate',
  'ftTime', 'ftDateTime', 'ftBytes', 'ftVarBytes', 'ftAutoInc', 'ftBlob', 'ftMemo',
  'ftGraphic', 'ftFmtMemo', 'ftParadoxOle', 'ftDBaseOle', 'ftTypedBinary', 'ftCursor');
{$ENDIF}*)

{$ELSE}
  FldType: array[TFieldType] of string = ('ftUnknown', 'ftString', 'ftSmallint',
   'ftInteger', 'ftWord', 'ftBoolean', 'ftFloat', 'ftCurrency', 'ftBCD', 'ftDate',
   'ftTime', 'ftDateTime', 'ftBytes', 'ftVarBytes', 'ftBlob', 'ftMemo', 'ftGraphic');
{$ENDIF}

begin
  FDataLink.FieldName := Value;
  if Field <> nil then
    if not (FDataLink.Field.DataType in [ftString, ftBoolean, ftSmallInt,
      ftInteger, ftWord, ftFloat, ftCurrency, ftDate, ftTime, ftDateTime]) then
        MessageDlg(Format('You cannot filter on %s field types. Please try again.',
          [FldType[FDataLink.Field.DataType]]), mtInformation, [mbOK], 0);
end;

{ get fieldname field }
function TDBFilter.GetField: TField;
begin
  Result := FDataLink.Field;
end;

{ set filter field }
procedure TDBFilter.SetFilter(Value: string);
begin
  if Value <> FFilter then
  begin
    FFilter := Value;
    FilterToBuf;
  end;
end;

procedure TDBFilter.ApplyFilter(Value: boolean);
begin
  if Value then
  begin
    AddFilter;
    FApply := true;
  end else
  begin
    RemoveFilter;
    FApply := false;
  end;
end;

procedure TDBFilter.FilterToBuf;
var
  I: integer;
  X: Double;
  B: boolean;
  D: LongInt;
begin
  if Field <> nil then
  begin
    FillChar(Buf, SizeOf(Buf), #0);
    case Field.DataType of
      ftString: Move(FFilter[1], Buf, Length(FFilter));
      ftBoolean: begin
        if Uppercase(FFilter) = 'TRUE' then B := true;
        if Uppercase(FFilter) = 'FALSE' then B := false;
        Move(B, Buf, SizeOf(B));
      end;
      ftSmallint, ftInteger,
      ftWord: begin
        I := StrToInt(FFilter);
        Move(I, Buf, SizeOf(I));
      end;
      ftFloat, ftCurrency,
      ftTime, ftDateTime: begin
        X := StrToFloat(FFilter);
        Move(X, Buf, SizeOf(X));
      end;
      ftDate: begin
        D := Trunc(StrToDate(FFilter));
        Move(D, Buf, SizeOf(LongInt));
      end;
    end;
  end;
end;

{ notify if DataSource removed }
procedure TDBFilter.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

{ set filter on table }
{$IFDEF Win32}

procedure TDBFilter.AddFilter;
type
  { Setup the node structure}
  TNodes = record
    BNode: CANBinary;
    FNode: CANField;
    CNode: CANConst;
  end;
var
  Nodes: TNodes;
  Expression: CANExpr;
  pCan: pByte;
  hFilter: hDbiFilter; { handle to filter }
begin
  { Binary Node - set the operator to canEQ }
  with Nodes.BNode do
  begin
    nodeClass := nodeBINARY;
    canOp := OpTypeMap[FOperator];
    iOperand1 := 12;
    iOperand2 := 24;
  end;

  { Field Node - set the field number and literal pool offset }
  with Nodes.FNode do
  begin
    nodeClass := nodeFIELD;
    canOp := canFIELD2;
    iFieldNum := Field.Index + 1;
    iNameOffset := 0;
  end;

  { Const Node - set the constant and literal pool offset }
  with Nodes.CNode do
  begin
    nodeClass := nodeCONST;
    canOp := canCONST2;
    iType := FldTypeMap[Field.DataType];
    iSize := SizeOf(Buf);
    iOffset := Length(Field.FieldName) + 1;
  end;

  { Expression - set the expression size, nodes and start positions }
  with Expression do
  begin
    iVer := 1;
    iTotalSize := sizeof(CANExpr) + sizeof(Nodes) + Length(Field.FieldName) + 1 + SizeOf(Buf);
    iNodes := 3;
    iNodeStart := sizeof(CANExpr);
    iLiteralStart := sizeof(CANExpr) + sizeof(Nodes);
  end;

  GetMem(pCan, Expression.iTotalSize * sizeof(BYTE));
  FillChar(pCan^, Expression.iTotalSize * sizeof(BYTE), #0);

  try
    { Move expression, nodes and literal pool into a contiguous memory space }
    Move(Expression, pCan^, sizeof(CANExpr));
    Inc(pCan, sizeof(CANExpr));
    Move(Nodes, pCan^, sizeof(Nodes));
    Inc(pCan, sizeof(Nodes));
    Move(Field.FieldName[1], pCan^, Length(Field.FieldName));
    Inc(pCan, Length(Field.FieldName) + 1);
    Move(Buf, pCan^, SizeOf(Buf));
    Dec(pCan, sizeof(Nodes) + sizeof(CANExpr) + Length(Field.FieldName) + 1);
    { Add and activate the filter }
    Check(DbiAddFilter((DataSource.DataSet as TTable).Handle, 0, 1, False, pCANExpr(pCan), nil, hFilter));
    Check(DbiActivateFilter((DataSource.DataSet as TTable).Handle, hFilter));
    (DataSource.DataSet as TTable).Refresh;
  finally
    FreeMem(pCan, Expression.iTotalSize * sizeof(BYTE));
  end;
end;

{$ELSE}

procedure TDBFilter.AddFilter;
{ custom type - stores filter expression and components }
type
  TMyFilter = record
    Expr: CANExpr;
    Nodes: array[0..2] of CANNode;
    Literals: array[0..255] of char;
  end;
var
  MyFilter: TMyFilter;  { instance of custom type }
  fldName: pChar; { fieldname char's }
  dbResult: DbiResult; { result from BDE }
  hFilter: hDbiFilter; { handle to filter }
begin
  if Field <> nil then
    try
      { allocate resources }
      fldName := StrAlloc(256);
      { copy }
      StrPCopy(fldName, DataField);

      { set the CANExpr field of the filter expression }
      MyFilter.Expr.iVer := 1;
      MyFilter.Expr.iTotalSize := sizeOf(MyFilter);
      MyFilter.Expr.iNodes := 3;
      MyFilter.Expr.iNodeStart := sizeOf(CANExpr);
      MyFilter.Expr.iLiteralStart := sizeOf(CANExpr) + (3 * sizeOf(CANNode));

      { set the Nodes field of the filter expression }
      { first part of array }
      MyFilter.nodes[0].canBinary.NodeClass  := nodeBinary;
      MyFilter.nodes[0].canBinary.canOP := OpTypeMap[FOperator];
      MyFilter.nodes[0].canBinary.iOperand1 := sizeOf(CANNode);
      MyFilter.nodes[0].canBinary.iOperand2 := 2 * sizeOf(CANNode);

      { second part of array }
      MyFilter.nodes[1].canField.nodeClass := nodeField;
      MyFilter.nodes[1].canField.canOP := canField2;
      MyFilter.nodes[1].canField.iFieldNum := 0;
      MyFilter.nodes[1].canField.iNameOffset := 0;

      { third part of array }
      MyFilter.nodes[2].canConst.nodeClass := nodeConst;
      MyFilter.nodes[2].canConst.canOP := canCONST2;
      MyFilter.nodes[2].canConst.iType := FldTypeMap[Field.DataType];
      MyFilter.nodes[2].canConst.iSize := 3;
      MyFilter.nodes[2].canConst.iOffset := StrLen(fldName) + 1;

      { add to array }
      { set query expression }
      FillChar(MyFilter.Literals, SizeOf(MyFilter.Literals), #0); { fill with #0 }
      StrCopy(MyFilter.Literals, fldName);      { set array to field-name }
      Move(Buf, MyFilter.Literals[StrLen(fldName)+1], SizeOf(Buf));

      { run filter }
      if DataSource.DataSet is TTable then
        dbResult := DBiAddFilter((DataSource.DataSet as TTable).Handle,
          0, 1, False, addr(MyFilter), nil, hFilter);
      if dbResult = DBIERR_NONE then
      begin
        dbResult := DBiActivateFilter((DataSource.DataSet as TTable).Handle, hFilter);
        if dbResult = DBIERR_NONE then
          DataSource.DataSet.First
        else MessageDlg('Could not activate filter', mtWarning, [mbOK], 0);
      end
      else MessageDlg('Could not set filter', mtWarning, [mbOK], 0);
    finally
      StrDispose(fldName);    { free resources allocated to fieldname }
    end;
end;
{$ENDIF}

{ remove filter }
procedure TDBFilter.RemoveFilter;
var
  dbResult: DBiResult; { result from BDE }
begin
  { if filter handle is null, then all table filters are deactivated }
  dbResult := DBiDeactivateFilter((DataSource.DataSet as TTable).Handle, nil);
  if dbResult = DBIERR_NONE then
  begin
    dbResult := DBiDropFilter((DataSource.DataSet as TTable).Handle, nil);
    if dbResult = DBIERR_NONE then DataSource.DataSet.First
    else MessageDlg('Could not dispose of filter', mtWarning, [mbOK], 0);
  end else
    MessageDlg('Could not deactivate filter', mtWarning, [mbOK], 0);
  (DataSource.DataSet as TTable).Refresh;
end;

{ register component on data controls page }
procedure Register;
begin
  RegisterComponents('Data Controls', [TDBFilter]);
end;

end.
